'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var debug$1 = require('debug');
var fs = require('fs');
var path = require('path');
var makeDir = require('make-dir');
var signale = require('signale');
var url = require('url');
var isMatch = require('lodash.ismatch');
var ciInfo = require('ci-info');
var crypto = require('crypto');
var sproutData = require('sprout-data');
var queryString = require('query-string');

let logLevel = process.env.DEBUG_LL || 'info';
let resultLogLevel = 'debug';
if (logLevel === 'debug') {
    resultLogLevel = 'info';
}
if (logLevel === 'info') {
    resultLogLevel = 'debug';
}
const options = {
    disabled: false,
    interactive: false,
    logLevel: resultLogLevel,
    scope: 'teremock',
    secrets: [],
    stream: process.stdout,
    types: {
        debug: {
            badge: '•',
            color: 'grey',
            label: 'debug',
            logLevel: 'info',
        },
        info: {
            badge: 'ø',
            color: 'green',
            label: 'info',
            logLevel: 'debug',
        },
        get: {
            badge: '»',
            color: 'gray',
            label: 'storage.get',
            logLevel: 'info',
        },
        set: {
            badge: 'œ',
            color: 'red',
            label: 'storage.set',
            logLevel: 'warn',
        },
    },
};
const logger$5 = new signale.Signale(options);
const loggersMap = {};
function debug(key) {
    if (!loggersMap[key]) {
        loggersMap[key] = debug$1(key);
    }
    return loggersMap[key];
}

const os = require('os');
const pathToUnixPath = os.platform() === 'win32' ? (str) => str.replace(/\\/g, '/') : (str) => str;
const loggerGet = debug$1('teremock:storage:get');
const loggerSet = debug$1('teremock:storage:set');
const loggerHas = debug$1('teremock:storage:has');
const getFileName = ({ wd, mockId }) => {
    return pathToUnixPath(path.resolve(wd, mockId.replace('--', path.sep) + '.json'));
};
class FileStorage {
    constructor(params) {
        this.wd = params?.wd ?? path.resolve(process.cwd(), '__teremocks__');
    }
    async set(mockId, json) {
        loggerSet(`entering storage.set with mockId "${mockId}"`);
        const absFileName = getFileName({ mockId, wd: this.wd });
        const targetDir = path.dirname(absFileName);
        const content = JSON.stringify(json, null, '  ');
        await makeDir(targetDir);
        loggerSet(`successfully checked/created targetDir ${targetDir}`);
        if (await this.has(mockId)) {
            loggerSet(`file already exists, overwriting ${mockId}`);
        }
        logger$5.set(`writing mock for mockId "${mockId}", url "${json?.request?.url}"`);
        logger$5.set(`to file "${absFileName}"`);
        try {
            await fs.promises.writeFile(absFileName, content);
            logger$5.success(`successfully wrote new file ${absFileName}`);
        }
        catch (e) {
            logger$5.error(`failed to write new file ${absFileName}`);
            throw e;
        }
    }
    // only for tests
    _getFn(mockId) {
        return getFileName({ mockId, wd: this.wd });
    }
    async get(mockId) {
        const absFileName = this._getFn(mockId);
        loggerGet(`about to read file ${absFileName}`);
        try {
            const data = JSON.parse(await fs.promises.readFile(absFileName, 'utf8'));
            loggerGet(`successfully read the file ${absFileName}`);
            return data;
        }
        catch (err) {
            if (err.code === 'ENOENT') {
                loggerGet(`fail to read the file (ENOENT) ${absFileName}`, err.message);
            }
            else {
                logger$5.error(`fail to read the file ${absFileName} for some other reason`, err);
            }
            throw err;
        }
    }
    async has(mockId) {
        const absFileName = this._getFn(mockId);
        loggerHas(`about to check file ${absFileName}`);
        let fileExists = false;
        try {
            await fs.promises.access(absFileName);
            fileExists = true;
            loggerHas(`file "${absFileName}" exists`);
        }
        catch (e) {
            loggerHas(`file "${absFileName}" does not exists`);
        }
        return fileExists;
    }
    setWd(nextWd) {
        this.wd = typeof nextWd === 'string' ? nextWd : path.join(process.cwd(), ...nextWd);
        return this.wd;
    }
}

// @ts-ignore
const WORDS = [
    'ack',
    'aim',
    'air',
    'and',
    'angel',
    'apart',
    'app',
    'april',
    'arizona',
    'ark',
    'artist',
    'aspen',
    'august',
    'autumn',
    'avocado',
    'axe',
    'bacon',
    'ban',
    'bark',
    'batman',
    'beam',
    'beer',
    'berlin',
    'biz',
    'black',
    'blue',
    'boo',
    'bravo',
    'bud',
    'bump',
    'burger',
    'bus',
    'butter',
    'carbon',
    'cardinal',
    'carolina',
    'carpet',
    'cat',
    'ceiling',
    'charlie',
    'chicken',
    'clip',
    'coffee',
    'cola',
    'cold',
    'color',
    'comet',
    'crazy',
    'cry',
    'cup',
    'cut',
    'dakota',
    'december',
    'delaware',
    'delta',
    'diet',
    'dog',
    'don',
    'door',
    'duck',
    'early',
    'earth',
    'east',
    'echo',
    'eco',
    'edward',
    'eight',
    'eme',
    'emma',
    'enemy',
    'equal',
    'failed',
    'fake',
    'fanta',
    'fifteen',
    'fillet',
    'finch',
    'fish',
    'five',
    'fix',
    'fiz',
    'floor',
    'flux',
    'fly',
    'four',
    'fruit',
    'gee',
    'georgia',
    'glucose',
    'goby',
    'golf',
    'green',
    'grey',
    'hamper',
    'happy',
    'harry',
    'hawaii',
    'helium',
    'high',
    'hook',
    'hot',
    'hotel',
    'hug',
    'hydrogen',
    'ice',
    'idaho',
    'illinois',
    'india',
    'indigo',
    'ink',
    'iowa',
    'island',
    'item',
    'jaws',
    'jazz',
    'jersey',
    'jet',
    'jig',
    'johnny',
    'juliet',
    'july',
    'jupiter',
    'kansas',
    'kentucky',
    'kid',
    'kilo',
    'king',
    'kitten',
    'lactose',
    'lake',
    'lamp',
    'lax',
    'lemon',
    'leopard',
    'lima',
    'limb',
    'lion',
    'lithium',
    'london',
    'low',
    'magazine',
    'maine',
    'mango',
    'march',
    'mars',
    'mass',
    'may',
    'mexico',
    'mic',
    'mike',
    'mirror',
    'miss',
    'mobile',
    'monkey',
    'moon',
    'muppet',
    'music',
    'neptune',
    'network',
    'nevada',
    'nine',
    'node',
    'north',
    'november',
    'nut',
    'nuts',
    'october',
    'ohio',
    'oil',
    'one',
    'orange',
    'oregon',
    'oscar',
    'oven',
    'owe',
    'oxid',
    'oxygen',
    'papa',
    'paris',
    'pasta',
    'perk',
    'pic',
    'pin',
    'pip',
    'pizza',
    'pluto',
    'potato',
    'queen',
    'quiet',
    'red',
    'rem',
    'ria',
    'river',
    'robin',
    'romeo',
    'rugby',
    'sad',
    'sad',
    'salami',
    'saturn',
    'seven',
    'sexy',
    'shade',
    'sierra',
    'single',
    'sink',
    'six',
    'sixteen',
    'skylark',
    'snake',
    'social',
    'sodium',
    'solar',
    'south',
    'spring',
    'spy',
    'steak',
    'steel',
    'stream',
    'summer',
    'sweet',
    'table',
    'tango',
    'tel',
    'ten',
    'tennis',
    'texas',
    'three',
    'tick',
    'timing',
    'tor',
    'toy',
    'twelve',
    'twenty',
    'two',
    'uncle',
    'uniform',
    'utah',
    'vegan',
    'venus',
    'victor',
    'video',
    'violet',
    'west',
    'white',
    'wiki',
    'winner',
    'winter',
    'xray',
    'yellow',
    'zag',
    'zebra',
    'zee',
    'zero',
    'zulu',
    'zuz',
];
const SEP = '-';
function humanize(hexdigest, words = 3) {
    if (typeof hexdigest !== 'string') {
        throw new Error(`hexdigest must be of type string`);
    }
    // if one char changed, all words must change
    const hashed = crypto.createHash('md5')
        .update(hexdigest)
        .digest('hex');
    const pairs = hashed.match(/(..?)/g);
    const bytes = pairs.map((x) => parseInt(x, 16));
    const compressed = compress(bytes, words);
    return compressed.map((x) => WORDS[x]).join(SEP);
}
function compress(bytes, len) {
    const length = bytes.length;
    if (len > length) {
        throw new Error('Fewer input bytes than requested output');
    }
    const seg_size = (length / len) >> 0;
    const segments = [];
    for (let i = 0; i < seg_size * len; i += seg_size) {
        segments.push(bytes.slice(i, i + seg_size));
    }
    segments[segments.length - 1] = segments[segments.length - 1].concat(bytes.slice(len * seg_size));
    const checksums = segments.map((x) => x.reduce((acc, curr) => acc ^ curr));
    return checksums;
}

// @ts-ignore
// Length of query string after which three-words-naming is switching on
const MAX_QUERY_NAME_LENGTH = 25;
function pickDeep(obj, keys) {
    return keys.reduce((acc, key) => {
        let val;
        if (Array.isArray(key)) {
            val = sproutData.get(obj, key);
            if (typeof val !== 'undefined') {
                sproutData.assoc(acc, key, val);
            }
        }
        else {
            val = sproutData.get(obj, [key]);
            sproutData.assoc(acc, [key], val);
        }
        return acc;
    }, {});
}
/**
 * Request Id generator.
 * Accounts for request data:
 * 1. url (protocol + host + path)
 * 2. query params (filtered with naming.query)
 * 3. method (get, post...) (as prefix)
 * 4. body, if any (filtered with naming.body)
 *
 * getRequestId does not accounts for
 * 1. blacklisted query and body params
 * 2. Any response data, including status, body and headers
 * 3. request headers (do you need that?)
 * 4. cookies
 *
 * Order agnostic, so, `/foo=bar&baz=1` will have the same rid as `/baz=1&foo=bar`
 */
const getRequestId = (params) => {
    let url$1 = params.url;
    let method = (params.method || 'GET').toLowerCase();
    let headers = params.headers;
    let body = params.body || '';
    let queryWhitelist = params.naming.query?.whitelist || [];
    let queryBlacklist = params.naming.query?.blacklist || [];
    let bodyWhitelist = params.naming.body?.whitelist || [];
    let bodyBlacklist = params.naming.body?.blacklist || [];
    const urlObj = new url.URL(url$1);
    let bodyObj;
    urlObj.searchParams.sort();
    // @todo remove it from here, rename body to body
    if (body !== '' && headers) {
        if (headers['content-type'] && headers['content-type'].includes('application/json')) {
            bodyObj = JSON.parse(body);
        }
        else {
            bodyObj = queryString.parse(body);
        }
    }
    if (bodyObj) {
        if (bodyWhitelist.length) {
            bodyObj = pickDeep(bodyObj, bodyWhitelist);
        }
        bodyBlacklist.forEach((param) => {
            let currentObj = bodyObj;
            let paramForDelete = param;
            if (Array.isArray(paramForDelete)) {
                const path = [...param];
                while (path.length > 1) {
                    // @ts-ignore
                    currentObj = currentObj[path.shift()];
                    if (!currentObj) {
                        return;
                    }
                }
                paramForDelete = path.shift();
            }
            delete currentObj[paramForDelete];
        });
        body = JSON.stringify(bodyObj);
    }
    /**
     * If naming.query.whitelist is set, removes any searchParams from the url,
     * which are not declared in the whitelist array
     */
    if (queryWhitelist.length > 0) {
        // @ts-ignore
        let keys = [...urlObj.searchParams.keys()];
        for (let key of keys) {
            if (!queryWhitelist.includes(key)) {
                urlObj.searchParams.delete(key);
            }
        }
    }
    // Some parameters could vary over the time, so we can exclude them from naming
    // (be carefull, use it only if it does not affect actual response body)
    // @ts-ignore
    queryBlacklist.forEach((param) => urlObj.searchParams.delete(param));
    let baseStr = urlObj.toString() + body;
    const entries = [];
    for (const [key, value] of urlObj.searchParams) {
        entries.push(`${key}-${value}`);
    }
    const queryStr = entries.join('-');
    let queryStrAscii = queryStr.toLowerCase().replace(/[^a-z0-9-]/g, '');
    if (method === 'get' && !queryStr.length) {
        return method;
    }
    if (method === 'get' && queryStr.length < MAX_QUERY_NAME_LENGTH && queryStr.length === queryStrAscii.length) {
        return `${method}-${queryStrAscii}`;
    }
    return `${method}-${humanize(baseStr, 3)}`;
};

/**
 * Returns uniq mock id.
 * Warning! Two different real request may have one mock id,
 * because getRequestId does not account for blacklisted query and body params.
 * Also, two equal mocks may be stored in two different files, because of
 * different options.wd
 */
const getMockId = (params) => {
    const { url, naming, name, method, body, headers } = params;
    const { hostname, pathname } = new URL(url);
    let pathId = name;
    if (!pathId || pathId.startsWith('__')) {
        const slugs = pathname.split('/').filter((s) => !!s);
        slugs.unshift(hostname);
        pathId = slugs.join('-');
    }
    const requestId = getRequestId({ url, naming, method, body, headers });
    const mockId = pathId + '--' + requestId;
    return mockId;
};

const DEFAULT_INTERCEPTOR_CAPTURE = {
    name: '__teremock_buildin_capture',
    url: '*',
    methods: new Set('*'),
    pass: false,
    naming: {},
    resourceTypes: new Set(['xhr', 'fetch']),
};
const DEFAULT_INTERCEPTOR_PASS = {
    name: '__teremock_buildin_pass',
    url: '*',
    methods: new Set('*'),
    pass: true,
    naming: {},
    resourceTypes: new Set('*'),
};
const defaultInterceptors = {
    // You could overwrite default interceptors, but do you really need that? Tell me in issue!
    __teremock_buildin_capture: DEFAULT_INTERCEPTOR_CAPTURE,
    __teremock_buildin_pass: DEFAULT_INTERCEPTOR_PASS,
};
const DEFAULT_OPTIONS = {
    interceptors: defaultInterceptors,
    skipResponseHeaders: [
        'date',
        'expires',
        'last-modified',
        'x-powered-by',
        'etag',
        'cache-control',
        'content-length',
        'server',
        'sec-ch-ua',
        'sec-ch-ua-mobile',
    ],
    skipRequestHeaders: [
        'user-agent',
        'accept-language',
        'accept',
    ],
    ci: ciInfo.isCi,
    awaitConnectionsOnStop: false,
    getMockId,
    // You need default value, or next teremock.start() (without options.wd) will use old wd value
    wd: path.resolve(process.cwd(), '__teremocks__'),
};
const DEFAULT_RESP_HEADERS = {
    'Content-Type': 'application/json; charset=utf-8'
};

const loggerint = debug$1('teremock:utils:interceptor');
const loggerBody = debug$1('teremock:utils:bodymatched');
const loggerTrace = debug$1('teremock:trace');
let orderCounter = 0;
function assignResponse(response1, response2) {
    return {
        ...response1,
        ...response2,
        headers: {
            ...response1.headers,
            ...response2?.headers,
        },
    };
}
const getTimeStampWithStrictOrder = () => {
    return {
        timestamp: Date.now(),
        order: ++orderCounter,
    };
};
const isInterceptorMatched = (interceptor, request) => {
    const { url, method, resourceType = 'xhr' } = request;
    return Object.keys(interceptor).reduce((acc, key) => {
        const value = interceptor[key];
        // Nothing to compare === wildcard (for given `key`)
        if (typeof value === 'undefined' || value === '*' || ((value && value instanceof Set) && value.has('*'))) {
            return acc;
        }
        switch (key) {
            case 'url':
                return acc && url.includes(value);
            case 'methods':
                return acc && value.has(method.toLowerCase());
            case 'query':
                const query = getQuery(url);
                return (acc &&
                    Object.keys(value).reduce((a, k) => {
                        return a && value[k] === query[k];
                    }, true));
            case 'body':
                return acc && isBodyMatched(value, request);
            case 'resourceTypes':
                return acc && value.has(resourceType.toLowerCase());
            default:
                return acc;
        }
    }, true);
};
const findInterceptor = ({ interceptors, request }) => {
    const matchedMockKey = Object.keys(interceptors).find((key) => {
        loggerint(`checking key`, key);
        const interceptor = interceptors[key];
        return isInterceptorMatched(interceptor, request);
    });
    loggerint(`found an interceptor`, matchedMockKey);
    return typeof matchedMockKey === 'string' ? interceptors[matchedMockKey] : null;
};
// duplicates are not supported
function getQuery(url$1) {
    const urlObj = new url.URL(url$1);
    urlObj.searchParams.sort();
    const entries = [...urlObj.searchParams];
    return entries.reduce((acc, [key, value]) => {
        acc[key] = value;
        return acc;
    }, {});
}
function isBodyMatched(value, request) {
    const { headers } = request;
    if (!headers) {
        return false;
    }
    let formData = {};
    if (headers['content-type'] === 'application/x-www-form-urlencoded') {
        formData = getFormData(request.body);
    }
    else if (headers['content-type'] === 'application/json'
        && typeof request.body === 'string') {
        try {
            formData = JSON.parse(request.body);
        }
        catch (e) {
            return false;
        }
    }
    // If body in UserInterceptor is empty - is considered equal
    if (!Object.keys(value).length) {
        return true;
    }
    const objectsIsMatch = isMatch(formData, value);
    if (!objectsIsMatch) {
        loggerBody('Body not matched. Request body: ', formData);
    }
    return objectsIsMatch;
}
// problem with decoding string, where spaces are encoded as "+" - it is correct for specification
// x-www-form-urlencoded content type https://www.rfc-editor.org/rfc/rfc1866,
// but JS decodeUri function doesn't decode it. same function exist in third-party libraries:
// https://chromium.googlesource.com/chromium/src.git/+/62.0.3178.1/third_party/google_input_tools/third_party/closure_library/closure/goog/string/string.js?autodive=0%2F%2F%2F%2F#486
function decodeUriWithSpaces(encodedStr) {
    return decodeURIComponent(encodedStr.replace(/\+/g, ' '));
}
function getFormData(body) {
    const result = {};
    if (typeof body === 'string') {
        const params = body.split('&');
        params.forEach(param => {
            const paramsPart = param.split('=');
            result[decodeUriWithSpaces(paramsPart[0])] = decodeUriWithSpaces(paramsPart[1]);
        });
    }
    return result;
}
function parseUrl(url$1) {
    const { origin, pathname } = new url.URL(url$1);
    return {
        query: getQuery(url$1),
        location: origin + pathname,
    };
}
function blacklist(source, list) {
    if (typeof source === 'undefined') {
        return source;
    }
    const set = new Set(list.map(l => l.toLowerCase()));
    return Object.keys(source).reduce((acc, key) => {
        if (!set.has(key.toLowerCase())) {
            acc[key] = source[key];
        }
        return acc;
    }, {});
}
function userInterceptorToInterceptor(uint, nameArg) {
    const name = uint.name || nameArg || `teremock-add-${humanize(JSON.stringify(uint), 1)}-`;
    const validName = name.toLowerCase().replace(/[^a-z0-9_-]+/, '');
    if (name.toLowerCase() !== validName) {
        throw new Error(`invalid interceptor name "${name}"! only letters, digits, - and _ are allowed.`);
    }
    const defaultMethods = DEFAULT_INTERCEPTOR_CAPTURE.methods;
    const defaultResourceTypes = DEFAULT_INTERCEPTOR_CAPTURE.resourceTypes;
    const methods = typeof uint.methods === 'string' ? new Set(uint.methods.toLowerCase().split(',')) : defaultMethods;
    const resourceTypes = typeof uint.resourceTypes === 'string' ? new Set(uint.resourceTypes.toLowerCase().split(',')) : defaultResourceTypes;
    let resp = {};
    if ('response' in uint && typeof uint.response !== 'function') {
        resp = {
            response: {
                url: '<unknown>',
                status: 200,
                ...uint.response,
            },
        };
    }
    return {
        ...DEFAULT_INTERCEPTOR_CAPTURE,
        ...uint,
        ...resp,
        methods,
        resourceTypes,
        name: validName,
    };
}
function userOptionsToOptions(defaultOptions, userOptions) {
    const defaultInterceptors = defaultOptions.interceptors;
    const { page, ...restUO } = userOptions;
    let interceptors = defaultInterceptors;
    if (typeof restUO.interceptors !== 'undefined') {
        const userInterceptors = restUO.interceptors;
        const customInterceptors = Object.keys(userInterceptors).reduce((acc, key) => {
            acc[key] = userInterceptorToInterceptor(userInterceptors[key], key);
            return acc;
        }, {});
        // Default interceptors must be in the end, but it should be possible to overwrite them
        interceptors = { ...customInterceptors, ...defaultInterceptors, ...customInterceptors };
    }
    let getMockId$1 = defaultOptions.getMockId;
    if (typeof userOptions.getMockId === 'function') {
        getMockId$1 = (...args) => {
            return userOptions?.getMockId?.(...args) ?? getMockId(...args);
        };
    }
    return {
        ...defaultOptions,
        ...restUO,
        interceptors,
        getMockId: getMockId$1,
    };
}
const getBody = (body) => {
    if (typeof body === 'string') {
        return body;
    }
    if (body == null) {
        return undefined;
    }
    return JSON.stringify(body);
};
class AsyncPendingQueue {
    constructor() {
        this.allSettledPromise = Promise.resolve();
        this._errorsCount = 0;
    }
    add(promise) {
        const catchedPromise = promise.catch(() => {
            this._errorsCount += 1;
        });
        this.allSettledPromise = Promise.all([this.allSettledPromise, catchedPromise]).then(() => null);
        return promise;
    }
    async awaitPending() {
        await this.allSettledPromise;
        return this._errorsCount;
    }
}

// const sleep = (time) => new Promise((resolve) => setTimeout(resolve, time))
const logger$4 = debug('teremock:request');
async function sleep(time) {
    return new Promise((resolve, _reject) => {
        setTimeout(resolve, time);
    });
}
async function beforeRespond({ respond, request, response, mog, interceptor, increment, responseOverrides, }) {
    let partResp;
    if (typeof response === 'function') {
        const argRequest = {
            ...request,
            query: getQuery(request.url),
        };
        mog(`» response is a function, responding with its returns`);
        partResp = await response(argRequest);
        mog(`» response() returns`, partResp);
    }
    else {
        partResp = response;
    }
    const bodyStrOrUnd = getBody(partResp.body);
    mog('» responding with', (bodyStrOrUnd ?? '').slice(0, 100));
    // @ts-ignore
    const resp = {
        ...partResp,
        headers: {
            ...DEFAULT_RESP_HEADERS,
            ...partResp.headers,
        },
        body: bodyStrOrUnd,
    };
    let resultResponse = assignResponse(resp, responseOverrides);
    const ttfb = resultResponse?.ttfb ?? 0;
    const actualDelay = Array.isArray(ttfb) ? ttfb[increment() % ttfb.length] : ttfb;
    mog('» actualDelay', actualDelay);
    await sleep(Math.floor(actualDelay));
    if (resultResponse.status >= 300 && resultResponse.status < 400) {
        const { body, ...rest } = resultResponse;
        resultResponse = rest;
    }
    await respond(resultResponse, interceptor);
}
function createHandler$1(initialParams) {
    let i = 0;
    const increment = () => i++;
    logger$4('creating request handler');
    return async function handleRequest({ request, abort, next, respond }, extraParams = {}) {
        const params = { ...initialParams, ...extraParams };
        const { interceptors, storage, reqSet, ci, responseOverrides, getMockId, _onMatch } = params;
        const reqParams = { url: request.url, method: request.method, body: request.body, headers: request.headers };
        loggerTrace(`${request.url} ← handling request`);
        logger$4(`» intercepted request with method "${request.method}" and url "${request.url}"`);
        logger$4(`request handling for:`, reqParams);
        logger$4(`request headers :`, request.headers);
        const interceptor = findInterceptor({ interceptors, request });
        if (!interceptor) {
            logger$5.warn(`interceptor not found, request to "${request.url}" will be aborted`);
            logger$5.warn(`it is recommended to add interceptors to cover all request from your test app`);
            logger$5.warn(`see https://github.com/diokuz/teremock#interceptor for details`);
            logger$4('» interceptor not found, aborting');
            await abort('aborted');
            return;
        }
        _onMatch(interceptor, request);
        const mockId = getMockId({ ...request, naming: interceptor.naming ?? {}, name: interceptor.name, body: getBody(request.body) });
        const mog = debug(`teremock:${mockId}`);
        mog(`interceptor found`, interceptor);
        if (interceptor.pass) {
            loggerTrace(`${request.url} ← passing to real server`);
            mog(`» interceptor.pass is true, sending it to real server next(interceptor)`);
            await next(interceptor);
            return;
        }
        params._onReqStarted(request);
        reqSet.add(mockId);
        mog('» reqSet is', Array.from(reqSet.get()));
        if (interceptor.response) {
            loggerTrace(`${request.url} ← inline response`);
            mog(`» interceptor.response defined, responding with it`);
            await beforeRespond({
                request,
                response: interceptor.response,
                responseOverrides,
                respond,
                interceptor,
                mog,
                increment,
            });
            return;
        }
        // mocks from storage
        mog(`» trying to get mock with id "${mockId}"`);
        if (await storage.has(mockId)) {
            mog(`» mock "${mockId}" exists!`);
            const mock = await storage.get(mockId);
            loggerTrace(`${request.url} ← mock ${mockId} found in storage`);
            mog(`» successfully read from "${mockId}", responding`);
            await beforeRespond({
                request,
                response: mock.response,
                responseOverrides,
                respond,
                interceptor,
                mog,
                increment,
            });
        }
        else {
            loggerTrace(`${request.url} ← mock not found in storage`);
            mog(`» mock does not exist!`, ci);
            if (ci) {
                logger$5.warn(`mock file not found in ci mode, url is "${request.url}"`);
            }
            else {
                mog('» about to next()...');
                await next(interceptor);
            }
        }
    };
}

const logger$3 = debug('teremock:response');
function createHandler(initialParams) {
    logger$3('creating response handler');
    return async function handleResponse({ request, response: pResponse, __meta }, extraParams = {}) {
        const params = { ...initialParams, ...extraParams };
        const { storage, reqSet, ci, skipResponseHeaders, skipRequestHeaders, getMockId } = params;
        loggerTrace(`${request.url} → handling response`);
        const response = {
            ...pResponse,
            headers: blacklist(pResponse.headers, skipResponseHeaders)
        };
        logger$3(`» intercepted response with method "${request.method}" and url "${request.url}"`);
        const interceptor = __meta?.interceptor;
        if (!interceptor) {
            loggerTrace(`${request.url} → no interceptor found :(`);
            logger$3(`» interceptor not provided in response object, return`);
            return;
        }
        const naming = interceptor.naming ?? {};
        const mockId = getMockId({ ...request, naming, name: interceptor.name, body: getBody(request.body) });
        const mockExist = await storage.has(mockId);
        const hasResp = !!interceptor.response;
        const mog = debug(`teremock:${mockId}`);
        /**
         * @todo update mock in non-ci mode
         */
        if (ci) {
            loggerTrace(`${request.url} → ci is true, not storing`);
            mog(`« mock was not stored because it is CI mode run`);
        }
        else if (mockExist) {
            loggerTrace(`${request.url} → mock already exists`);
            mog(`« mock was not stored because it exists`);
        }
        else if (hasResp) {
            loggerTrace(`${request.url} → interceptor.response is defined`);
            mog(`« mock was not stored because matched interceptor have response property`);
        }
        else if (interceptor.pass) {
            loggerTrace(`${request.url} → interceptor.pass is true, no store`);
            mog(`« mock was not stored because interceptor.pass is true`);
        }
        else {
            loggerTrace(`${request.url} → storing mock ${mockId}`);
            mog(`« preparing to set a new mock "${mockId}"`);
            const { timestamp: _x, order: _z, id, ...mockRequest } = request;
            const { timestamp: _y, order: _k, ...mockResponse } = response;
            if (mockRequest.headers) {
                mockRequest.headers = blacklist(mockRequest.headers, skipRequestHeaders);
            }
            await storage.set(mockId, { request: mockRequest, response: mockResponse });
        }
        reqSet.delete(mockId);
        mog('« reqSet after delete is', Array.from(reqSet));
        params._onReqCompleted(request, response);
        if (reqSet.size === 0) {
            mog('« invoking _onReqsCompleted');
            params._onReqsCompleted();
        }
    };
}

const logger$2 = debug$1('teremock:driver:request');
let requestsCounter = 0;
function extractPlaywrightRequest(route, options) {
    const playwrightRequest = route.request();
    const request = {
        url: playwrightRequest.url(),
        method: playwrightRequest.method(),
        headers: playwrightRequest.headers(),
        body: playwrightRequest.postData(),
        resourceType: playwrightRequest.resourceType(),
        id: requestsCounter++,
        timestamp: options.timestamp,
        order: options.order,
    };
    // @ts-expect-error
    playwrightRequest.timestamp = options.timestamp;
    // @ts-expect-error
    playwrightRequest.teremockRequest = request;
    // @ts-expect-error
    playwrightRequest.__meta = { request };
    logger$2(`got the request ${request.url}, sending it to teremock core`);
    return {
        request,
        abort: async (errorCode) => await route.abort(errorCode),
        next: async (interceptor) => {
            logger$2(`continue() call`);
            // @ts-expect-error
            playwrightRequest.__meta.interceptor = interceptor;
            loggerTrace(`${request.url} ← request.continue()`);
            await route.continue();
        },
        respond: async (response, interceptor) => {
            logger$2(`respond() call`);
            // @ts-expect-error
            playwrightRequest.__meta.interceptor = interceptor;
            loggerTrace(`${request.url} ← request.respond(...)`);
            await route.fulfill(response);
        },
    };
}

const logger$1 = debug$1('teremock:driver:response');
async function extractPlaywrightResponse(playwrightResponse, options) {
    const playwrightRequest = playwrightResponse.request();
    let requestBody;
    let responseBody;
    if (options.noParseResponse) {
        responseBody = '';
        requestBody = '';
    }
    else {
        try {
            responseBody = await playwrightResponse.json();
        }
        catch (e) {
            logger$1('« playwrightResponse.json() error:', e.message);
            try {
                responseBody = await playwrightResponse.text();
                logger$1(`« response body starts with: ${responseBody?.substr(0, 20)}`);
            }
            catch (e) {
                logger$1('« playwrightResponse.text() error:', e.message);
            }
        }
        try {
            requestBody = JSON.parse(playwrightRequest.postData());
        }
        catch (e) {
            requestBody = playwrightRequest.postData();
        }
    }
    // @ts-expect-error
    const { teremockRequest } = playwrightRequest;
    const request = {
        url: playwrightRequest.url(),
        method: playwrightRequest.method(),
        headers: playwrightRequest.headers(),
        body: requestBody,
        resourceType: playwrightRequest.resourceType(),
        id: teremockRequest ? teremockRequest.id : -1,
        timestamp: teremockRequest ? teremockRequest.timestamp : 0,
        order: teremockRequest ? teremockRequest.order : -1,
    };
    const response = {
        url: playwrightResponse.url(),
        status: playwrightResponse.status(),
        headers: playwrightResponse.headers(),
        body: responseBody,
        timestamp: options.timestamp,
        order: options.order,
        // ttfb: timestamp - playwrightRequest.timestamp,
    };
    logger$1(`got the response, sending it to teremock core`);
    const { __meta } = playwrightRequest;
    if (!__meta) {
        logger$5.warn(`__meta was not found in playwrightRequest. Probably it was made before teremock.start()`);
        logger$5.warn(`Passing the response witout storing it. The request was:`, request);
    }
    return { request, response, __meta };
}

// https://playwright.dev/docs/network
/**
 * There is no valid reason to have more than one driver instances per page
 */
const pagesSet = new Set();
const redirectCodes = [301, 302, 303, 305, 307, 308];
class PlaywrightDriver {
    constructor({ page, noParseResponseUrls = [] }) {
        this._routeUrl = () => true;
        logger$5.debug(`instantiating new playwright driver`);
        if (!page) {
            throw new Error('"page" is not defined');
        }
        if (pagesSet.has(page)) {
            logger$5.error(`second driver instantiation on the same page`);
            logger$5.error(`(probably you did not stop the mocker before start it again)`);
            throw new Error(`second driver instantiation on the same page`);
        }
        this.seenRedirects = new Map();
        this.noParseResponseUrls = noParseResponseUrls;
        this.page = page;
        pagesSet.add(page);
        this.page.on('close', () => {
            pagesSet.delete(this.page);
        });
    }
    async setRequestInterception(_arg) {
        // do nothing for playwright
    }
    async onRequest(fn) {
        const handler = async (route) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            loggerTrace(`${route.request().url()} ← page.on('request') fired`);
            const { request, abort, next, respond } = extractPlaywrightRequest(route, timestampWithOrder);
            await fn({ request, abort, next, respond });
        };
        await this.page.route(this._routeUrl, handler);
        return async () => {
            pagesSet.delete(this.page);
            try {
                await this.page.unroute(this._routeUrl, handler);
            }
            catch (e) {
                if (e.message.indexOf('has been closed') === -1) {
                    throw e;
                }
            }
        };
    }
    async onResponse(fn) {
        const handler = async (interceptedResponse) => {
            const timestampWithOrder = getTimeStampWithStrictOrder();
            const url$1 = interceptedResponse.request().url();
            if (redirectCodes.indexOf(interceptedResponse.status()) !== -1) {
                // see https://playwright.dev/docs/api/class-page#pagerouteurl-handler
                // Because of the way the page.route method works, the following redirection will not be handled by it.
                const newLocation = interceptedResponse.headers().location;
                const destUrl = new url.URL(newLocation, url$1).toString();
                this.seenRedirects.set(destUrl, (this.seenRedirects.get(destUrl) || 0) + 1);
                logger$5.debug(`Redirection to "${newLocation}" from "${url$1}" will not be handled.`);
            }
            if (this.seenRedirects.get(url$1)) {
                this.seenRedirects.set(url$1, this.seenRedirects.get(url$1) - 1);
                return;
            }
            const noParseResponse = this.noParseResponseUrls.some((urlPart) => url$1.indexOf(urlPart) !== -1);
            loggerTrace(`${url$1} → page.on('response') fired`);
            await fn(await extractPlaywrightResponse(interceptedResponse, {
                ...timestampWithOrder,
                noParseResponse,
            }));
            loggerTrace(`${url$1} → finish`);
        };
        // Intercepting all requests and responding with mocks
        await this.page.on('response', handler);
        return () => this.page.off('response', handler);
    }
}

/*
 * It is a singleton, the only instance is located in `exports`
 * @todo instantiate for each test suite explicitly
 */
const logger = debug$1('teremock');
function noop() { }
class Teremock {
    constructor(opts) {
        logger$5.debug(`new Mocker`);
        this.defaultOptions = Object.assign({}, DEFAULT_OPTIONS);
        this.options = this.defaultOptions;
        this.alive = false;
        this._spies = [];
        this._interceptors = {};
        this._resolveReqs = noop;
        this._matched = new Map();
        this.storage = opts?.storage ?? new FileStorage();
        this.driver = opts?.driver;
        this.reqSet = new Set();
        this.reqSetRet = new Set();
    }
    _onAnyReqStart(req) {
        if (this.reqSet.size === 0) {
            this.reqsPromise = new Promise((resolve) => {
                this._resolveReqs = () => {
                    resolve(Array.from(this.reqSetRet));
                    this.reqSetRet.clear();
                };
            });
        }
        if (this._spies.length > 0) {
            this._spies.forEach(([interceptor, spy]) => {
                if (isInterceptorMatched(interceptor, req)) {
                    spy.called = true;
                    spy.callCount++;
                    spy.calledOnce = spy.callCount === 1;
                    if (req.id !== -1 && req.id && req.order && req.timestamp) {
                        spy.events.push({
                            requestTimestamp: req.timestamp,
                            requestOrder: req.order,
                            requestId: req.id
                        });
                    }
                }
            });
        }
    }
    _onAnyReqEnd(req, resp) {
        if (this._spies.length > 0) {
            this._spies.forEach(([interceptor, spy]) => {
                if (isInterceptorMatched(interceptor, req)) {
                    if (req.id !== -1 && req.id) {
                        const eventInfo = spy.events.find(obj => obj.requestId === req.id);
                        if (eventInfo) {
                            eventInfo.responseTimestamp = resp.timestamp;
                            eventInfo.responseOrder = resp.order;
                        }
                    }
                }
            });
        }
    }
    /*
     * Starts to intercept requests
     */
    async start(userOptions) {
        logger$5.debug(`mocker.start()`);
        if (this.alive) {
            logger$5.warn(`mocker was not stopped before new start call, stopping it first`);
            await this.stop();
        }
        let resolveStartPromise;
        this._startPromise = new Promise((resolve) => {
            resolveStartPromise = resolve;
        });
        this.alive = true;
        this.options = userOptionsToOptions(this.defaultOptions, userOptions);
        this.options.wd && this.storage.setWd?.(this.options.wd);
        this.driver = this.driver ?? new PlaywrightDriver({ page: userOptions.page });
        await this.driver.setRequestInterception(true);
        this._spies = [];
        this._interceptors = this.options.interceptors;
        this._matched = new Map();
        const logParams = Object.assign({}, this.options);
        logger('Mocker starts with resulting params:', logParams);
        const pureRequestHandler = createHandler$1({
            ...this.options,
            interceptors: this._interceptors,
            storage: this.storage,
            reqSet: {
                add: (en) => {
                    this.reqSet.add(en);
                    this.reqSetRet.add(en);
                },
                get: () => this.reqSet,
            },
            _onReqStarted: (req) => this._onAnyReqStart(req),
            _onMatch: (interceptor, req) => {
                if (!this._matched.has(interceptor.name)) {
                    this._matched.set(interceptor.name, []);
                }
                this._matched.get(interceptor.name).push(req.url);
            }
        });
        this._requestPendingQueue = new AsyncPendingQueue();
        const pureResponseHandler = createHandler({
            ...this.options,
            interceptors: this._interceptors,
            storage: this.storage,
            reqSet: this.reqSet,
            _onReqCompleted: (req, resp) => this._onAnyReqEnd(req, resp),
            _onReqsCompleted: () => this._resolveReqs(),
        });
        this._responsePendingQueue = new AsyncPendingQueue();
        logger('Handlers created, params validated');
        // Intercepting all requests and respinding with mocks
        this.removeRequestHandler = await this.driver.onRequest(async (ir) => {
            // driver may calls onRequest callback synchronously,
            // we using here AsyncPendingQueue who can track pending promises
            try {
                await this._requestPendingQueue.add(pureRequestHandler(ir, {
                    ...this.extraParams,
                    interceptors: this._interceptors,
                }));
            }
            catch (e) {
                logger$5.error(`There is error in request handler:`);
                logger$5.error(e);
            }
        });
        // Writing mocks on real responses to filesystem
        this.removeResponseHandler = await this.driver.onResponse(async (ir) => {
            // driver may calls onResponse callback synchronously,
            // we using here AsyncPendingQueue who can track pending promises
            try {
                await this._responsePendingQueue.add(pureResponseHandler(ir, {
                    ...this.extraParams,
                    interceptors: this._interceptors,
                }));
            }
            catch (e) {
                logger$5.error(`There is error in response handler:`);
                logger$5.error(e);
            }
        });
        logger('_startPromise about to resolve (Request interception enabled, listeners added)');
        resolveStartPromise();
        return this._startPromise;
    }
    add(userInterceptor, overwrite = false) {
        const interceptor = userInterceptorToInterceptor(userInterceptor);
        const { name } = interceptor;
        if (name in this._interceptors && !overwrite) {
            logger$5.error(`interceptor with name "${name}" already exists, pass second arg true if you want to overwrite it`);
            return () => { };
        }
        this._interceptors = {
            [name]: interceptor,
            ...this._interceptors,
        };
        // @todo support restoring overwritten names
        return () => {
            const { [name]: omit, ...rest } = this._interceptors;
            this._interceptors = rest;
        };
    }
    remove(userInterceptor) {
        const interceptor = userInterceptorToInterceptor(userInterceptor);
        const { name } = interceptor;
        if (!this._interceptors[name]) {
            logger$5.error(`there is no interceptor with name "${name}"`);
        }
        const { [name]: _x, ...rest } = this._interceptors;
        this._interceptors = rest;
    }
    /*
     * Resolves when all mocked connections are completed
     * @deprecated
     */
    connections(opts = { timeout: 1000 }) {
        logger$5.warn(`teremock.connections() is deprecated`);
        logger$5.warn(`it is recommended to await explicit effects as a result of requests`);
        const { timeout } = opts;
        if (typeof this._startPromise === 'undefined') {
            throw new Error('Cant await connections. Probably you didnt start the mocker?');
        }
        const timeoutPromise = new Promise((_resolve, reject) => {
            setTimeout(() => reject(new Error(`failed to await connectins in ${timeout} ms`)), timeout);
        });
        if (typeof this.reqsPromise === 'undefined') {
            throw new Error(`no connections, nothing to await`);
        }
        return Promise.race([timeoutPromise, this.reqsPromise]);
    }
    spy(spyInterceptor) {
        const interceptor = userInterceptorToInterceptor(spyInterceptor, 'spy');
        const spy = {
            callCount: 0,
            calledOnce: false,
            called: false,
            dismiss: () => {
                this._spies = this._spies.filter(([_i, s]) => s !== spy);
            },
            events: [],
        };
        this._spies.push([interceptor, spy]);
        return spy;
    }
    /*
     * Waits for all connections to be completed and removes all handlers from the page
     */
    async stop(opts = {}) {
        const { safe } = opts;
        if (!this.alive) {
            if (safe) {
                return;
            }
            // Async, because jest suppress Errors in `after` callbacks
            setTimeout(() => {
                logger$5.warn(`nothing to stop. Did you call 'mocker.stop()' twice?`);
            });
            return;
        }
        const t1 = setTimeout(() => {
            if (this.reqSet.size === 0) {
                logger(`Mocker failed to stop. reqSet.size === 0`);
            }
            else {
                logger(`Mocker failed to stop. Alive connections:\n${[...this.reqSet]}`);
            }
            logger('About to throw an Error');
            logger$5.error(`Failed to stop mocker!`);
            logger$5.error(`Possible reasons:`);
            logger$5.error(`1) navigation happened before all responses completed.`);
            logger$5.error(`2) some connections are not finished in 20 seconds.`);
            logger$5.error(`The number of opened connections is ${this.reqSet.size}.\nReqSet:${[...this.reqSet]}`);
            throw new Error(`Failed to stop mocker!`);
        }, 20 * 1000);
        logger('Begining of mocker.stop procedure');
        if (typeof this._startPromise === 'undefined') {
            throw new Error('Cant stop mocker. Probably you didnt start it?');
        }
        let failed = false;
        await this._startPromise;
        if (this.options.awaitConnectionsOnStop) {
            try {
                await this.connections();
            }
            catch (err) {
                logger$5.error(`this.connections was rejected with error. Continue...`, err);
                failed = err;
            }
        }
        logger('`Start` promise was resolved, `connections` promise was resolved or rejected');
        clearTimeout(t1);
        const requestHandlerErrorsCount = await this._requestPendingQueue?.awaitPending();
        if (requestHandlerErrorsCount) {
            throw new Error(`There were ${requestHandlerErrorsCount} errors in the request handler. See error output above.`);
        }
        const responseHandlerErrorsCount = await this._responsePendingQueue?.awaitPending();
        if (responseHandlerErrorsCount) {
            throw new Error(`There were ${responseHandlerErrorsCount} errors in the response handler. See error output above.`);
        }
        await this.removeCloseHandler?.();
        await this.removeRequestHandler?.();
        await this.removeResponseHandler?.();
        this.alive = false;
        // @todo how to solve that without closure?
        // Must reject on `connections` reject, but must finish all tasks also
        if (failed) {
            logger$5.error(`rejecting mocker.stop() promise, because connections promise was rejected`);
            logger(`About to exit from mocker.stop with reject`);
            throw failed;
        }
        logger('Interceptor names and matched urls', this._matched);
        this.options.onStop?.({ matched: this._matched });
        logger(`about to exit from mocker.stop with resolve`);
    }
}

const teremock = new Teremock();

exports.AsyncPendingQueue = AsyncPendingQueue;
exports.PlaywrightDriver = PlaywrightDriver;
exports.Teremock = Teremock;
exports.assignResponse = assignResponse;
exports.blacklist = blacklist;
exports.default = teremock;
exports.findInterceptor = findInterceptor;
exports.getBody = getBody;
exports.getFormData = getFormData;
exports.getQuery = getQuery;
exports.getTimeStampWithStrictOrder = getTimeStampWithStrictOrder;
exports.isBodyMatched = isBodyMatched;
exports.isInterceptorMatched = isInterceptorMatched;
exports.loggerTrace = loggerTrace;
exports.parseUrl = parseUrl;
exports.userInterceptorToInterceptor = userInterceptorToInterceptor;
exports.userOptionsToOptions = userOptionsToOptions;
